//
// angle.cxx
// kkyoda@riken.jp
//

#define Pi 3.14159265

#include <memory>   // std::auto_ptr
#include <iostream>
#include <cstring>
#include <cmath>
#include <unistd.h>
#include <limits.h>
#include <list>
#include <sstream>

#include "bdml0.18.hxx"
#include "xyzt.h"

using namespace std;

int
main (int argc, char* argv[])
{
  try {
    if (argc < 6) {
      cerr << "usage: " << argv[0] << " -{id, name} 1st_component{ID, Name} 2nd_component{ID, Name} 3rd_component{ID, Name} *.bdml0.18.xml" << endl;
      return 1;
    }

    auto_ptr<bdml::bdmlDocument> h (bdml::bdml (argv[5]));

    for (int i = 6; i < argc; i++) {
      auto_ptr<bdml::bdmlDocument> r (bdml::bdml (argv[i]));
      if (r->data()) {
        for (bdml::Data::component_const_iterator p = r->data()->component().begin(); p != r->data()->component().end(); p++) {
          h->data()->component().push_back(*p);
        }
      }
    }


    int mode = 0;
    string comp1;
    string comp2;
    string comp3;
    string smode = argv[1];
    if (!strcmp(smode.c_str(), "-id")) {
      mode = 1;
      comp1 = argv[2];
      comp2 = argv[3];
      comp3 = argv[4];
    } else if (!strcmp(smode.c_str(), "-name")) {
      mode = 2;
      comp1 = argv[2];
      comp2 = argv[3];
      comp3 = argv[4];
    } else {
      cerr << "usage: " << argv[0] << " -{id, name} 1st_component{ID, Name} 2nd_component{ID, Name} 3rd_component{ID, Name} *.bdml0.18.xml" << endl;
      return 1;
    }

    double gx;
    double gy;
    double gz;
    double area;
    double total;

    list<xyzt> xyztl1;
    list<xyzt> xyztl2;
    list<xyzt> xyztl3;
    xyztl1.clear();
    xyztl2.clear();
    xyztl3.clear();

    stringstream ss;
    int flag = 0;

    double xscale = 0.0;
    xscale = h->data()->scaleUnit().xScale();
    double yscale = 0.0;
    yscale = h->data()->scaleUnit().yScale();
    double zscale = 0.0;
    zscale = h->data()->scaleUnit().zScale();
    string xyzunit;
    xyzunit = h->data()->scaleUnit().xyzUnit();
    double tscale = 0.0;
    tscale = h->data()->scaleUnit().tScale();
    string tunit;
    tunit = h->data()->scaleUnit().tUnit();

    for (bdml::Data::component_const_iterator p = h->data()->component().begin(); p != h->data()->component().end(); p++) {
      int compflag = 0;
      if (mode == 1) {
        if (!strcmp(p->componentID().c_str(), comp1.c_str())) {
          compflag = 1;
        } else if (!strcmp(p->componentID().c_str(), comp2.c_str())) {
          compflag = 2;
        } else if (!strcmp(p->componentID().c_str(), comp3.c_str())) {
          compflag = 3;
        }
      } else if (mode == 2) {
        if (p->componentName()) {
          if (!strcmp(p->componentName()->c_str(), comp1.c_str())) {
            compflag = 1;
          } else if (!strcmp(p->componentName()->c_str(), comp2.c_str())) {
            compflag = 2;
          } else if (!strcmp(p->componentName()->c_str(), comp3.c_str())) {
            compflag = 3;
          }
        }
      }

      if ((compflag == 1) | (compflag == 2) | (compflag == 3)) {
        gx = gy = gz = area = total = 0.0;
        for (bdml::Component::measurement_const_iterator r = p->measurement().begin(); r != p->measurement().end(); r++) {
          if (r->point()) {
            for (bdml::Point::xyz_const_iterator q = r->point()->xyz().begin(); q != r->point()->xyz().end(); q++) {
              double x, y, z;
              x = q->x();
              y = q->y();
              z = q->z();
              gx += x;
              gy += y;
              gz += z;
              total += 1.0;
            }
            flag = 1;
          } else if (r->line()) {
            for (bdml::Line::xyzSequence_const_iterator q = r->line()->xyzSequence().begin(); q != r->line()->xyzSequence().end(); q++) {
              double x = 0.0, y = 0.0, z = 0.0;
              double prev_x = 0.0, prev_y = 0.0;
              int sflag = 0;
              double startx = 0.0, starty = 0.0, startz = 0.0;
              int cnt = 0;

              for (bdml::XYZSequence::xyz_const_iterator s = q->xyz().begin(); s != q->xyz().end(); s++) {
                cnt++;
                prev_x = x;
                prev_y = y;
                x = s->x();
                y = s->y();
                z = s->z();
                if (sflag == 0) {
                  startx = x;
                  starty = y;
                  startz = z;
                  sflag = 1;
                }
              }

              if ((startx == x) && (starty == y) && (startz == z)) {
                double subarea = 0.0;
                for (bdml::XYZSequence::xyz_const_iterator s = q->xyz().begin(); s != q->xyz().end(); s++) {
                  x = s->x();
                  y = s->y();
                  z = s->z();
                  gx += (prev_x + x) * (prev_x * y - x * prev_y);
                  gy += (prev_y + y) * (prev_x * y - x * prev_y);
                  subarea += prev_x * y - x * prev_y;

                  prev_x = x;
                  prev_y = y;
                }
                subarea = subarea / 2.0;
                area += subarea;
                gz += z * subarea; 

                flag = 2;
              } else {
                for (bdml::XYZSequence::xyz_const_iterator s = q->xyz().begin(); s != q->xyz().end(); s++) {
                  x = s->x();
                  y = s->y();
                  z = s->z();
                  gx += x;
                  gy += y;
                  gz += z;
                  total += 1.0;
                }

                flag = 3;
              }
            }

          } else if (r->face()) {
            double dv = 0.0;
            double vol = 0.0;
            double dx = 0.0, dy = 0.0, dz = 0.0;
            double dmomx = 0.0, dmomy = 0.0, dmomz = 0.0;
            for (bdml::Face::xyzTriplet_const_iterator q = r->face()->xyzTriplet().begin(); q != r->face()->xyzTriplet().end(); q++) {
              double ax = 0.0, ay = 0.0, az = 0.0, bx = 0.0, by = 0.0, bz = 0.0, cx = 0.0, cy = 0.0, cz = 0.0;
              bdml::XYZTriplet::xyz_const_iterator s = q->xyz().begin();
              ax = s->x() * xscale;
              ay = s->y() * yscale;
              az = s->z() * zscale;
              s++;
              bx = s->x() * xscale;
              by = s->y() * yscale;
              bz = s->z() * zscale;
              s++;
              cx = s->x() * xscale;
              cy = s->y() * yscale;
              cz = s->z() * zscale;

              dv = (ax * (by * cz - bz * cy) + ay * (bz * cx - bx * cz) + az * (bx * cy - by * cx)) / 6.0;
              vol += dv;

              dx = (ax + bx + cx) / 4.0;
              dy = (ay + by + cy) / 4.0;
              dz = (az + bz + cz) / 4.0;

              dmomx += dx * dv;
              dmomy += dy * dv;
              dmomz += dz * dv;
            }

            gx = dmomx / vol;
            gy = dmomy / vol;
            gz = dmomz / vol;

            flag = 4;

          } else if (r->circle()) {
            double x, y, z;
            double rad;
            x = r->circle()->xyz().x();
            y = r->circle()->xyz().y();
            z = r->circle()->xyz().z();
            rad = r->circle()->radius();
            gx += x * Pi * rad * rad;
            gy += y * Pi * rad * rad;
            gz += z * Pi * rad * rad;
            total += Pi * rad * rad;

            flag = 5;

          } else if (r->sphere()) {
            double x, y, z;
            double rad;
            x = r->sphere()->xyz().x();
            y = r->sphere()->xyz().y();
            z = r->sphere()->xyz().z();
            rad = r->sphere()->radius();
            gx += x * 4.0 * Pi * rad * rad / 3.0;
            gy += y * 4.0 * Pi * rad * rad / 3.0;
            gz += z * 4.0 * Pi * rad * rad / 3.0;
            total += 4.0 * Pi * rad * rad / 3.0;

            flag = 6;
          }
        }

        double xx, yy, zz;
        if (flag == 1) {  // point
          xx = gx / total * xscale;
          yy = gy / total * yscale;
          zz = gz / total * zscale;

          xyzt tmp_xyzt(xx, yy, zz, p->time());
          if (compflag == 1)
            xyztl1.push_back(tmp_xyzt);
          else if (compflag == 2)
            xyztl2.push_back(tmp_xyzt);
          else if (compflag == 3)
            xyztl3.push_back(tmp_xyzt);
        } else if (flag == 2) {  // line (closed)
          xx = gx / 6.0 / area * xscale;
          yy = gy / 6.0 / area * yscale;
          zz = gz / area * zscale;

          xyzt tmp_xyzt(xx, yy, zz, p->time());
          if (compflag == 1)
            xyztl1.push_back(tmp_xyzt);
          else if (compflag == 2)
            xyztl2.push_back(tmp_xyzt);
          else if (compflag == 3)
            xyztl3.push_back(tmp_xyzt);
        } else if (flag == 3) {  // line
          xx = gx / total * xscale;
          yy = gy / total * yscale;
          zz = gz / total * zscale;

          xyzt tmp_xyzt(xx, yy, zz, p->time());
          if (compflag == 1)
            xyztl1.push_back(tmp_xyzt);
          else if (compflag == 2)
            xyztl2.push_back(tmp_xyzt);
          else if (compflag == 3)
            xyztl3.push_back(tmp_xyzt);
        } else if (flag == 4) {  // face
          xx = gx;
          yy = gy;
          zz = gz;

          xyzt tmp_xyzt(xx, yy, zz, p->time());
          if (compflag == 1)
            xyztl1.push_back(tmp_xyzt);
          else if (compflag == 2)
            xyztl2.push_back(tmp_xyzt);
          else if (compflag == 3)
            xyztl3.push_back(tmp_xyzt);
        } else if (flag == 5) {  // circle
          xx = gx / total * xscale;
          yy = gy / total * yscale;
          zz = gz / total * zscale;

          xyzt tmp_xyzt(xx, yy, zz, p->time());
          if (compflag == 1)
            xyztl1.push_back(tmp_xyzt);
          else if (compflag == 2)
            xyztl2.push_back(tmp_xyzt);
          else if (compflag == 3)
            xyztl3.push_back(tmp_xyzt);
        } else if (flag == 6) {  // sphere
          xx = gx / total * xscale;
          yy = gy / total * yscale;
          zz = gz / total * zscale;

          xyzt tmp_xyzt(xx, yy, zz, p->time());
          if (compflag == 1)
            xyztl1.push_back(tmp_xyzt);
          else if (compflag == 2)
            xyztl2.push_back(tmp_xyzt);
          else if (compflag == 3)
            xyztl3.push_back(tmp_xyzt);
        } else {  // other
          cout << "this program can handle only Point, Line, Circle, and Sphere structure." << endl;
          exit(-1);
        }
      }
    }

    xyztl1.sort();
    list<xyzt>::iterator v;
    xyztl2.sort();
    list<xyzt>::iterator u;
    xyztl3.sort();
    list<xyzt>::iterator w;

    if ((xyztl1.size() == 0) | (xyztl2.size() == 0) | (xyztl3.size() == 0)) {
      cerr << "no component: " << comp1 << " or " << comp2 << " or " << comp3 << endl;
      exit(-1);
    }

    int prev_t1 = -1000;
    int prev_t2 = -1000;
    int prev_t3 = -1000;
    for (v = xyztl1.begin(), u = xyztl2.begin(), w = xyztl3.begin(); v != xyztl1.end(), u != xyztl2.end(), w != xyztl3.end(); v++, u++, w++) {
      if ((prev_t1 == v->getT()) | (prev_t2 == u->getT()) | (prev_t3 == w->getT())) {
        cerr << "more than one component for " << comp1 << ", " << comp2 << ", or " << comp3 << " exist in the same time point" << endl;
        exit(-1);
      }
      prev_t1 = v->getT();
      prev_t2 = u->getT();
      prev_t3 = w->getT();
    }

    int eflag = 0;
    for (v = xyztl1.begin(); v != xyztl1.end(); v++) {
      for (u = xyztl2.begin(); u != xyztl2.end(); u++) {
        for (w = xyztl3.begin(); w != xyztl3.end(); w++) {
          if ((v->getT() == u->getT()) && (v->getT() == w->getT())) {
            eflag = 1;
          }
        }
      }
    }
    if (eflag == 0) {
      cerr << "three components, " << comp1 << ", " << comp2 << ", and " << comp3 << ", do not exist in the same time point" << endl;
      exit(-1);
    }

    if (mode == 1) {
      cout << "#componentID: " << comp1 << ", " << comp2 << ", and " << comp3 << ", Unit: " << "degree";
    } else if (mode == 2) {
      cout << "#componentName: " << comp1 << ", " << comp2 << ", and " << comp3 << ", Unit: " << "degree";
    }
    cout << ", time interval: " << tscale << " " << tunit << "(s)" << endl;
    cout << "#t\tangle" << endl;
    for (v = xyztl1.begin(); v != xyztl1.end(); v++) {
      for (u = xyztl2.begin(); u != xyztl2.end(); u++) {
        for (w = xyztl3.begin(); w != xyztl3.end(); w++) {
          if ((v->getT() == u->getT()) && (v->getT() == w->getT())) {
            double vec1x, vec1y, vec1z, vec2x, vec2y, vec2z;
            vec1x = (v->getX() - u->getX()) / sqrt((v->getX() - u->getX()) * (v->getX() - u->getX()) + (v->getY() - u->getY()) * (v->getY() - u->getY()) + (v->getZ() - u->getZ()) * (v->getZ() - u->getZ()));
            vec1y = (v->getY() - u->getY()) / sqrt((v->getX() - u->getX()) * (v->getX() - u->getX()) + (v->getY() - u->getY()) * (v->getY() - u->getY()) + (v->getZ() - u->getZ()) * (v->getZ() - u->getZ()));
            vec1z = (v->getZ() - u->getZ()) / sqrt((v->getX() - u->getX()) * (v->getX() - u->getX()) + (v->getY() - u->getY()) * (v->getY() - u->getY()) + (v->getZ() - u->getZ()) * (v->getZ() - u->getZ()));
            vec2x = (w->getX() - u->getX()) / sqrt((w->getX() - u->getX()) * (w->getX() - u->getX()) + (w->getY() - u->getY()) * (w->getY() - u->getY()) + (w->getZ() - u->getZ()) * (w->getZ() - u->getZ()));
            vec2y = (w->getY() - u->getY()) / sqrt((w->getX() - u->getX()) * (w->getX() - u->getX()) + (w->getY() - u->getY()) * (w->getY() - u->getY()) + (w->getZ() - u->getZ()) * (w->getZ() - u->getZ()));
            vec2z = (w->getZ() - u->getZ()) / sqrt((w->getX() - u->getX()) * (w->getX() - u->getX()) + (w->getY() - u->getY()) * (w->getY() - u->getY()) + (w->getZ() - u->getZ()) * (w->getZ() - u->getZ()));
            double cost;
            cost = vec1x * vec2x + vec1y * vec2y + vec1z * vec2z;
            double radian;
            radian = acos(cost);
            double angle;
            angle = radian * 180.0 / Pi;
            cout << v->getT() << "\t" << angle << endl;
          }
        }
      }
    }
  }

  catch (const xml_schema::exception& e)
  {
    cerr << e << endl;
    return 1;
  }

}
