//
// count.cxx
// kkyoda@riken.jp
//

#define Pi 3.14159265

#include <memory>   // std::auto_ptr
#include <iostream>
#include <cstring>
#include <cmath>
#include <unistd.h>
#include <limits.h>
#include <list>

#include "bdml0.18.hxx"
#include "xyzt.h"

using namespace std;

int
main (int argc, char* argv[])
{

  try {
    if (argc < 4) {
      cerr << "usage: " << argv[0] << " -name componentName *.bdml0.18.xml" << endl;
      return 1;
    }

    auto_ptr<bdml::bdmlDocument> h (bdml::bdml (argv[3]));

    for (int i = 4; i < argc; i++) {
      auto_ptr<bdml::bdmlDocument> r (bdml::bdml (argv[i]));
      if (r->data()) {
        for (bdml::Data::component_const_iterator p = r->data()->component().begin(); p != r->data()->component().end(); p++) {
          h->data()->component().push_back(*p);
        }
      }
    }

    int mode = 0;
    string comp;
    string smode = argv[1];
    if (!strcmp(smode.c_str(), "-name")) {
      mode = 2;
      comp = argv[2];
    } else {
      mode = 1;
    }

    list<xyzt> xyztl;
    xyztl.clear();

    double xscale = 0.0;
    xscale = h->data()->scaleUnit().xScale();
    double yscale = 0.0;
    yscale = h->data()->scaleUnit().yScale();
    double zscale = 0.0;
    zscale = h->data()->scaleUnit().zScale();
    string xyzunit;
    xyzunit = h->data()->scaleUnit().xyzUnit();
    double tscale = 0.0;
    tscale = h->data()->scaleUnit().tScale();
    string tunit;
    tunit = h->data()->scaleUnit().tUnit();

    for (bdml::Data::component_const_iterator p = h->data()->component().begin(); p != h->data()->component().end(); p++) {
      int compflag = 0;
      if (mode == 2) {
        if (!strcmp(p->componentName()->c_str(), comp.c_str())) {
          compflag = 1;
        }
      } else if (mode == 1) {
        compflag = 1;
      }

      if (compflag == 1) {
          xyzt tmp_xyzt(1.0, 1.0, 1.0, p->time());
          xyztl.push_back(tmp_xyzt);
      }
    }

    xyztl.sort();
    list<xyzt>::iterator v;

    int st = 0;
    int et = 0;
    v = xyztl.begin();
    st = v->getT();
    for (;v != xyztl.end(); v++) {
      et = v->getT();
    }

    if (xyztl.size() == 0) {
      cerr << "no component: " << comp << endl;
      exit(-1);
    }

    if (mode == 2) {
      cout << "#componentName: " << comp << ", Unit: " << xyzunit;
    }
    cout << ", time interval: " << tscale << " " << tunit << "(s)" << endl;
    cout << "#t\tcount" << endl;
    int count = 0;
    for (int i = st; i <= et; i++) {
      count = 0;
      for (v = xyztl.begin(); v != xyztl.end(); v++) {
        if (v->getT() == i) {
          count++;
        }
      }
      cout << i << "\t" << count << endl;
    }
  }

  catch (const xml_schema::exception& e)
  {
    cerr << e << endl;
    return 1;
  }

}
