//
// timelength.cxx
// kkyoda@riken.jp
//

#include <memory>   // std::auto_ptr
#include <iostream>
#include <sstream>
#include <cstring>
#include <unistd.h>
#include <limits.h>
#include "bdml0.18.hxx"
using namespace std;

//stringstream ss;

int main(int argc, char* argv[])
{
  int mintime = +32767;
  int maxtime = -32768;

  try {
    if (argc < 3) {
      cerr << "usage: " << argv[0] << " componentName *.bdml0.18.xml" << endl;
      return 1;
    }

    auto_ptr<bdml::bdmlDocument> h (bdml::bdml (argv[2]));

    for (int i = 3; i < argc; i++) {
      auto_ptr<bdml::bdmlDocument> r (bdml::bdml (argv[i]));
      if (r->data()) {
        for (bdml::Data::component_const_iterator p = r->data()->component().begin(); p != r->data()->component().end(); p++) {
          h->data()->component().push_back(*p);
        }
      }
    }

    string compname = argv[1];

    double tscale;
    tscale = h->data()->scaleUnit().tScale();
    string tunit;
    tunit = h->data()->scaleUnit().tUnit();
    int flag = 0;

    for (bdml::Data::component_const_iterator p = h->data()->component().begin(); p != h->data()->component().end(); p++) {
      if (p->componentName()) {
        if (!strcmp(p->componentName()->c_str(), compname.c_str())) {
          cerr << p->time() << endl;
          if (mintime > p->time()) {
            mintime = p->time();
          } else if (maxtime < p->time()) {
            maxtime = p->time();
          }
          flag = 1;
        }
      }
    }

    if (flag == 1) {
      cout << "#componentName: " << compname << ", Unit: " << tunit << endl;
      cout << "#time length" << endl; 
      if ((maxtime - (mintime - 1)) * tscale == 1.0) {
        cout << (double) (maxtime - (mintime - 1)) * tscale << endl;
      } else {
        cout << (double) (maxtime - (mintime - 1)) * tscale << endl;
      }
    } else {
      cout << "no componentName: " << compname << endl;
    }
  }
  catch (const xml_schema::exception& e)
  {
    cerr << e << endl;
    return 1;
  }

}
