//
// volume.cxx
// kkyoda@riken.jp
//

#define Pi 3.14159265

#include <memory>   // std::auto_ptr
#include <iostream>
#include <cstring>
#include <cmath>
#include <unistd.h>
#include <limits.h>
#include <list>
#include <sstream>

#include "bdml0.18.hxx"
#include "vt.h"

using namespace std;

int
main (int argc, char* argv[])
{
  try {
    if (argc < 4) {
      cerr << "usage: " << argv[0] << " -{id, name} component{ID, Name} *.bdml0.18.xml" << endl;
      return 1;
    }

    auto_ptr<bdml::bdmlDocument> h (bdml::bdml (argv[3]));

    for (int i = 4; i < argc; i++) {
      auto_ptr<bdml::bdmlDocument> r (bdml::bdml (argv[i]));
      if (r->data()) {
        for (bdml::Data::component_const_iterator p = r->data()->component().begin(); p != r->data()->component().end(); p++) {
          h->data()->component().push_back(*p);
        }
      }
    }

    int mode = 0;
    string comp;
    string smode = argv[1];
    if (!strcmp(smode.c_str(), "-id")) {
      mode = 1;
      comp = argv[2];
    } else if (!strcmp(smode.c_str(), "-name")) {
      mode = 2;
      comp = argv[2];
    } else {
      cerr << "usage: " << argv[0] << " -{id, name} component{ID, Name} *.bdml0.18.xml" << endl;
      return 1;
    }

    double area;
    double vol;
    double length;
    double total;

    list<vt> vtl;
    vtl.clear();

    stringstream ss;
    int flag = 0;

    double xscale = 0.0;
    xscale = h->data()->scaleUnit().xScale();
    double yscale = 0.0;
    yscale = h->data()->scaleUnit().yScale();
    double zscale = 0.0;
    zscale = h->data()->scaleUnit().zScale();
    string xyzunit;
    xyzunit = h->data()->scaleUnit().xyzUnit();
    double tscale = 0.0;
    tscale = h->data()->scaleUnit().tScale();
    string tunit;
    tunit = h->data()->scaleUnit().tUnit();

    for (bdml::Data::component_const_iterator p = h->data()->component().begin(); p != h->data()->component().end(); p++) {
      int compflag = 0;
      if (mode == 1) {
        if (!strcmp(p->componentID().c_str(), comp.c_str())) {
          compflag = 1;
        }
      } else if (mode == 2) {
        if (p->componentName()) {
          if (!strcmp(p->componentName()->c_str(), comp.c_str())) {
            compflag = 1;
          }
        }
      }

      if (compflag == 1) {
        area = total = vol = length = 0.0;
        for (bdml::Component::measurement_const_iterator r = p->measurement().begin(); r != p->measurement().end(); r++) {
          if (r->point()) {
            flag = 1;
          } else if (r->line()) {
            for (bdml::Line::xyzSequence_const_iterator q = r->line()->xyzSequence().begin(); q != r->line()->xyzSequence().end(); q++) {
              double x = 0.0, y = 0.0, z = 0.0;
              double prev_x = 0.0, prev_y = 0.0;
              int sflag = 0;
              double startx = 0.0, starty = 0.0, startz = 0.0;
              int cnt = 0;
              for (bdml::XYZSequence::xyz_const_iterator s = q->xyz().begin(); s != q->xyz().end(); s++) {
                cnt++;
                prev_x = x;
                prev_y = y;
                x = s->x();
                y = s->y();
                z = s->z();
                if (sflag == 0) {
                  startx = x;
                  starty = y;
                  startz = z;
                  sflag = 1;
                }
              }

              if ((startx == x) && (starty == y) && (startz == z)) {
                double subarea = 0.0;
                for (bdml::XYZSequence::xyz_const_iterator s = q->xyz().begin(); s != q->xyz().end(); s++) {
                  x = s->x();
                  y = s->y();
                  z = s->z();
                  subarea += prev_x * y - x * prev_y;

                  prev_x = x;
                  prev_y = y;
                }
                subarea = subarea / 2.0;
                area += subarea;

                flag = 2;
              } else {
                flag = 3;
              }
            }

          } else if (r->face()) {
            double dv = 0.0;
            vol = 0.0;
            for (bdml::Face::xyzTriplet_const_iterator q = r->face()->xyzTriplet().begin(); q != r->face()->xyzTriplet().end(); q++) {
              double ax = 0.0, ay = 0.0, az = 0.0, bx = 0.0, by = 0.0, bz = 0.0, cx = 0.0, cy = 0.0, cz = 0.0;
              bdml::XYZTriplet::xyz_const_iterator s = q->xyz().begin();
              ax = s->x() * xscale;
              ay = s->y() * yscale;
              az = s->z() * zscale;
              s++;
              bx = s->x() * xscale;
              by = s->y() * yscale;
              bz = s->z() * zscale;
              s++;
              cx = s->x() * xscale;
              cy = s->y() * yscale;
              cz = s->z() * zscale;

              dv = (ax * (by * cz - bz * cy) + ay * (bz * cx - bx * cz) + az * (bx * cy - by * cx)) / 6.0;
              vol += dv;

            }

            flag = 4;

          } else if (r->circle()) {
            flag = 5;

          } else if (r->sphere()) {
            double rad = r->sphere()->radius();
            vol += 4.0 * Pi * rad * rad / 3.0;

            flag = 6;
          }
        }

        double val;
        if (flag == 1) {  // point
          cerr << "calculation was not executed: point" << endl;
          exit(-1);
        } else if (flag == 2) {  // line (closed)
          if (zscale != 0.0) {
            val = abs(area) * xscale * yscale * zscale;
          } else {
            cerr << "calculation was not executed: line (two-dimension)" << endl;
            exit(-1);
          }

          vt tmp_vt(val, p->time());
          vtl.push_back(tmp_vt);
        } else if (flag == 3) {  // line
          cerr << "calculation was not executed: line (not closed)" << endl;
          exit(-1);
        } else if (flag == 4) {  // face
          val = vol;
        } else if (flag == 5) {  // circle
          cerr << "calculation was not executed: circle" << endl;
          exit(-1);
        } else if (flag == 6) {  // sphere
          val = vol;

          vt tmp_vt(val, p->time());
          vtl.push_back(tmp_vt);
        } else {  // other
          cout << "this program can handle only Line (Closed) and Sphere structure." << endl;
          exit(-1);
        }
      }
    }

    vtl.sort();
    list<vt>::iterator v;

    if (vtl.size() == 0) {
      cerr << "no component: " << comp << endl;
      exit(-1);
    }

    if (mode == 1) {
      cout << "#componentID: " << comp;
    } else if (mode == 2) {
      cout << "#componentName: " << comp;
    }
    if (flag == 2) {
      cout << ", Unit: " << xyzunit << "^3, time interval: " << tscale << " " << tunit << "(s)" << endl;
      cout << "#t\tvolume" << endl;
      for (v = vtl.begin(); v != vtl.end(); v++) {
        cout << v->getT() << "\t" << v->getV() << endl;
      }
    } else if (flag == 6) {
      cout << ", Unit: " << xyzunit << "^3, time interval: " << tscale << " " << tunit << "(s)" << endl;
      cout << "#t\tvolume" << endl;
      for (v = vtl.begin(); v != vtl.end(); v++) {
        cout << v->getT() << "\t" << v->getV() << endl;
      }
    }
  }

  catch (const xml_schema::exception& e)
  {
    cerr << e << endl;
    return 1;
  }

}
